
__version__ = '0.4.12'

__authors__ = [
    ("Thomas G. Close", "tom.g.close@gmail.com"),
    ("Francesco Sforazzini", "francesco.sforazzini@gmail.com"),
    ("Phillip G. D. Ward", "phillipgdward@gmail.com")]

install_requires = [
    'xnat>=0.3.17',
    'nipype>=1.1.7',
    'pydicom>=1.0.2',
    'networkx>=2.2',
    'fasteners>=0.7.0',
    'future>=0.16.0',
    'pybids>=0.5.1',
    'contextlib2>=0.5.5',
    'deepdiff>=3.3']
